/* Virtuino example for ESP32 board without Virtuino library
 * Created by Ilias Lamprou
 * Updated Apr 29 2018
 * This code doesn't support DV pins. 
 * 
 * Download latest Virtuino android app from the link: https://play.google.com/store/apps/details?id=com.virtuino_automations.virtuino
 */
 
#include <ESP8266WiFi.h>

//------ USER SETTINGS ------------------------------------------------
const char* ssid = "WIFI NETWORK";            // WIFI network SSID
const char* password = "PASSWORD";            // WIFI network PASSWORD
WiFiServer wifiServer(8000);                  // Server port

//------ ENABLE THE LINE OF YOUR BOARD
#define WEMOS_D1_MINI_PRO                  //WeMos D1 mini or other with similar pin map  
// #define NODEMCU                         //NodeMCU  or other with similar pin map  
// #define ESP8266_UNSPECIFIED              //General ESP8266 module. You have to config the pinsMap manually
                                           
//-------PINS and MEMORY SETTINGS -------------------------------------

 #if defined WEMOS_D1_MINI_PRO                                 //-------- pinsMap for Wemos D1 mini board  
    #define boardPinsCount 9                                    // Board digital pins count
    const byte digitalPinsMap[]={D0,D1,D2,D3,D4,D5,D6,D7,D8};        // digital pinsMap
  #elif defined NODEMCU                                         //-------- pinsMap for NodeMCU board
    #define boardPinsCount 11                                   // Board digital pins count
    const byte digitalPinsMap[]={D0,D1,D2,D3,D4,D5,D6,D7,D8,D9,D10}; // digital pinsMap
 #else                                                          //-------- pinsMap for other esp boards
    #define boardPinsCount 11                                   // ESP general board pins count
    const byte digitalPinsMap[]={0,1,2,3,4,5,12,13,14,15,16};        // Board digital pins count
 #endif
  
  #define analogInputPinsCount 1                              // Analog pins count
  const byte analogInputPinsMap[]={A0};                            // analog pinsmap


//============================================================== connectToWiFiNetwork
void connectToWiFiNetwork(){
  Serial.println("Connecting to "+String(ssid));
  
   // If you don't want to config IP manually disable the next four lines
  IPAddress ip(192, 168, 1, 150);            // where 150 is the desired IP Address   
  IPAddress gateway(192, 168, 1, 1);         // set gateway to match your network
  IPAddress subnet(255, 255, 255, 0);        // set subnet mask to match your network
  WiFi.config(ip, gateway, subnet);          // If you don't want to config IP manually disable this line
  
  WiFi.mode(WIFI_STA);                       // Config module as station only.
  WiFi.begin(ssid, password);
   while (WiFi.status() != WL_CONNECTED) {
     delay(500);
     Serial.print(".");
    }
   Serial.println("");
   Serial.println("WiFi connected");
   Serial.println(WiFi.localIP());
}

//=================================================================
void initAccessPoint(){
  Serial.print("Setting soft-AP ... ");                   // Default IP: 192.168.4.1
   WiFi.mode(WIFI_AP);                                     // Config module as Access point only.  Set WiFi.mode(WIFI_AP_STA); to config module as Acces point and station
   boolean result = WiFi.softAP("Virtuino Network", "12345678");      // SSID: Virtuino Network   Password:12345678
   if(result == true)  {
    Serial.println("Server Ready");
    Serial.println(WiFi.softAPIP());
   }
   else Serial.println("Failed!");
}

//============================================================== setup
//==============================================================
void setup() {
  Serial.begin(9600);
  connectToWiFiNetwork();      //enable this line to connect the module your wifi network
  //initAccessPoint();         //enable this line to use the esp as AP  
  wifiServer.begin();
  pinMode(D4,OUTPUT);
  }

//============================================================== loop
//==============================================================
void loop() {
  virtuinoRun();

  //---- example to read a value from Virtuino 
  float v = readValueFromVirtualMemory(2);   // read pin V2
  if (v==1) digitalWrite(2,1); 
  else digitalWrite(2,0); 

  //---- example to write a value to Virtuino 
  int sensorValue=random(100);    // replace with your sensor value
  writeValueToVirtualMemory(4,sensorValue);   // write to  pin V4
  
}


//================== Virtuino Functions and Variables ==========
//============================================================== 
float virtualMemory[100];        // virtual memory V  -  size=100  This memory is synchronized with Virtuino app virtual memory V         
                                 // DV virtual memory doesn't supported on this sketch. DV memory is deprecated

#define REQUEST_PREFIX  "1234"                 // Virtuino password
#define ERROR_PREFIX "!E00=5"                  // Error prefix 
//============================================================== readVirtualMemoryValue
float readValueFromVirtualMemory(int pin){return virtualMemory[pin];}

//============================================================== writeVirtualMemoryValue
float writeValueToVirtualMemory(int pin, float value){virtualMemory[pin]=value;}

//============================================================== getEspPin
byte getEspDigitalPin(int  pin){
  if ((pin>=0) and (pin<boardPinsCount)) return digitalPinsMap[pin];
  else return  digitalPinsMap[0];
}
//============================================================== getEspPin
byte getEspAnalogPin(int  pin){
  if ((pin>=0) and (pin<analogInputPinsCount)) return analogInputPinsMap[pin];
  else return  analogInputPinsMap[0];
}

//============================================================== checkVirtuinoCommand
String checkVirtuinoCommand(char commandType, int pin, String valueAsText){
  int isQuestionCommand = valueAsText.indexOf('?');
  String pinString = String(pin);
  if (pin<10) pinString = "0"+pinString;
  
    else pinString=String(pin);
  if (commandType=='V') {                       // this example works only with V memory 
     if (isQuestionCommand==-1) virtualMemory[pin]=valueAsText.toFloat();
     return "!V"+String(pin)+'='+String(virtualMemory[pin])+'$';  // response
     }
  else if (commandType=='A') {                       
     int v=analogRead(getEspAnalogPin(pin));      
     return "!A"+pinString+'='+String(v)+'$';  // response
     }   
  else if (commandType=='Q') {                       
       if (isQuestionCommand==-1) {
        byte v=1;
        if (valueAsText.toInt()==0) v=0;
        digitalWrite(getEspDigitalPin(pin),v);
        }
      return "!Q"+pinString+'='+String(digitalRead(getEspDigitalPin(pin)))+'$';  // response
      }
   else if (commandType=='C') return "!C00=1.70$";  // connected- Return a number as firmware version
        
     
  return '!'+String(commandType)+pinString+"=0$";    // for zero for all other commands 
}
//============================================================== 
String createResponseForVirtuino(String* commands){
   String commandResponse="";
  // read all virtuino commands from request
  boolean checkNext=true;
  int cStart=-1;
  while (checkNext){
    cStart=commands->indexOf('!',cStart+1);
    if (cStart!=-1) {
      int cEnd=commands->indexOf('$',cStart);
      String oneCommand=commands->substring(cStart+1,cEnd); 
      char commandType = oneCommand.charAt(0);
      int equalCharPosition = oneCommand.indexOf('=');
      int pin = oneCommand.substring(1,equalCharPosition).toInt(); 
      String valueString = oneCommand.substring(equalCharPosition+1); 
      commandResponse+=checkVirtuinoCommand(commandType,pin,valueString);
    }
    else checkNext=false;   // the last command has checked
  }
  
  return commandResponse;
}

//============================================================== virtuinoRun
void   virtuinoRun(){
    WiFiClient client = wifiServer.available();   // Check if a client has connected
    if (client) {
      while(!client.available())delay(1);    // Wait until the client sends some data
      String req = client.readStringUntil('\r');   // Read the first line of the request
      Serial.println(req);
      

      //-- check request
      int pos = req.indexOf("GET /");
      if (pos!=-1){                                   // We have a GET message
        req.remove(0,pos+5);        // Clear the characters GET / 
        String commandResponse="";
        //-- check prefix           // default prefix is 1234. Use this prefix to distinguish requests from different clients
       if (req.startsWith(REQUEST_PREFIX))commandResponse = createResponseForVirtuino(&req); 
         else commandResponse=ERROR_PREFIX;
        String responseHttp= "HTTP/1.1 200 OK\r\n Content-Type: text/html\r\nConnection: close\r\n\r\n";
        responseHttp+=commandResponse;       // add the commands response to request response
        responseHttp+="\r\n\r\n";
        Serial.println("responseHttp="+responseHttp);
        client.print(responseHttp);
        delay(1);
        Serial.println("Client disonnected");
        client.flush();
    }
    else {
      Serial.println("invalid request");
      client.stop();
    }
}
}


